import numpy as np
import scipy.linalg
import joblib

from .tools import median_mad




def extract_chunks(signals, left_sample_indexes, width, 
                channel_indexes=None, channel_adjacency=None, 
                peak_channel_indexes=None, chunks=None):
    """
    This cut small chunks on signals and return concatenate them.
    This use numpy.array for input/output.
    
    
    Arguments
    ---------------
    signals: np.ndarray 
        shape (nb_campleXnb_channel)
    left_sample_indexes: np.ndarray
        sample postion of the first sample
    width: int
        Width in sample of chunks.
    channel_indexes: 
        Force a limited number of channels. If None then all channel.
    
    channel_adjacency: dict
        dict of channel adjacency
        If not None, then the wavzform extraction is sprase given adjency.
        In that case peak_channel_indexes must be provide and channel_indexes 
        must be None.
    peak_channel_indexes: np.array None
        Position of the peak on channel space used only when channel_adjacency not None.
        
    Returns
    -----------
    chunks : np.ndarray
        shape = (left_sample_indexes.size, width, signals.shape[1], )
    
    """
    if channel_adjacency is not None:
        assert channel_indexes is None
        assert peak_channel_indexes is not None, 'For sparse eaxtraction peak_channel_indexes must be provide'
    
    if chunks is None:
        if channel_indexes is None:
            chunks = np.empty((left_sample_indexes.size, width, signals.shape[1]), dtype = signals.dtype)
        else:
            chunks = np.empty((left_sample_indexes.size, width, len(channel_indexes)), dtype = signals.dtype)
        
    keep = (left_sample_indexes>=0) & (left_sample_indexes<(signals.shape[0] - width))
    left_sample_indexes2 = left_sample_indexes[keep]
    if peak_channel_indexes is not None:
        peak_channel_indexes2 = peak_channel_indexes[keep]
    
    if channel_indexes is None and channel_adjacency is None:
        # all channels
        for i, ind in enumerate(left_sample_indexes2):
            chunks[i,:,:] = signals[ind:ind+width,:]
    elif channel_indexes is not None:
        for i, ind in enumerate(left_sample_indexes2):
            chunks[i,:,:] = signals[ind:ind+width,:][:, channel_indexes]
            
    elif channel_adjacency is not None:
        # sparse
        assert peak_channel_indexes is not None
        for i, ind in enumerate(left_sample_indexes2):
            chan = peak_channel_indexes2[i]
            chans = channel_adjacency[chan]
            chunks[i,:,:][:, chans] = signals[ind:ind+width,:][:, chans]

    return chunks



def equal_template_with_distance(centroid0, centroid1, thresh=2.0, n_shift = 2):
    """
    Test template equality based on distance between centroids.
    Do this for some jitter (n_shift)
    
    """
    wf0 = centroid0[n_shift:-n_shift, :]
    
    equal = False
    for shift in range(n_shift*2+1):
        wf1 = centroid1[shift:wf0.shape[0]+shift, :]
        
        d = np.max(np.abs(wf1 - wf0))

        #~ print('shift', shift, 'd', d)

        #~ import matplotlib.pyplot as plt
        #~ fig, ax = plt.subplots()
        #~ ax.plot(centroid0.T.flatten())
        #~ ax.plot(centroid1.T.flatten())
        #~ ax.set_title(f'shift {shift} thresh {thresh:0.2f} d {d:0.2f} merge ' + str(d<thresh))
        #~ plt.show()
        
        if d<thresh:
            equal = True
                
            #~ import matplotlib.pyplot as plt
            #~ fig, ax = plt.subplots()
            #~ ax.plot(centroid0.T.flatten())
            #~ ax.plot(centroid1.T.flatten())
            #~ ax.set_title(f'shift {shift} thresh {thresh:0.2f} d {d:0.2f} merge ' + str(d<thresh))
            #~ plt.show()
            
            break
    
    return equal



def equal_template_with_distrib_overlap(centroid0, waveforms0, centroid1, waveforms1,
                    n_shift = 2, quantile_limit=0.8, sparse_thresh=1.5,
                    debug=False
                    ):
    """
    Test template equality with checking distribution overlap.
    Do this for some jitter (n_shift)
    
    """
    mask0 = np.any(np.abs(centroid0) > sparse_thresh, axis=0)
    mask1 = np.any(np.abs(centroid1) > sparse_thresh, axis=0)
    mask = mask1 | mask0
    
    #~ print(centroid0.shape)

    #~ centroid0_ = centroid0[np.newaxis, n_shift:-n_shift, :]
    
    centroid0_ = centroid0[n_shift:-n_shift, :]
    centroid1_ = centroid1[n_shift:-n_shift, :]
    vector_0_1 = (centroid1_ - centroid0_)
    vector_0_1 =vector_0_1[:, mask]
    vector_0_1 /= np.sum(vector_0_1**2)
    

    
    wfs0 = waveforms0[:, n_shift:-n_shift, :].copy()
    #~ dist_0_to_0 = np.sum((centroid0_[:,:, mask] - wfs0[:,:, mask]) ** 2, axis=(1,2))
    #~ print(dist_0_to_0)
    
    #~ limit = np.quantile(dist_0_to_0, quantile_limit)

    equal = False
    for shift in range(n_shift*2+1):
        #~ print('shift', shift)
        wfs1 = waveforms1[:, shift:centroid0_.shape[0]+shift, :].copy()
        
        inner_sp = np.sum((wfs0[:,:,mask] - centroid0_[np.newaxis,:,:][:,:,mask]) * vector_0_1[np.newaxis,:,:], axis=(1,2))
        cross_sp = np.sum((wfs1[:,:,mask] - centroid0_[np.newaxis,:,:][:,:,mask]) * vector_0_1[np.newaxis,:,:], axis=(1,2))
        
        l0 = np.quantile(inner_sp, quantile_limit)
        l1 = np.quantile(cross_sp, 1 - quantile_limit)
        
        equal = l0 >= l1
        

        
        #~ if np.median(dist_1_to_0)<limit:
        #~ if True:
        #~ if equal:
        if debug:
                
            import matplotlib.pyplot as plt

            fig, ax = plt.subplots()
            count, bins = np.histogram(inner_sp, bins=100)
            ax.plot(bins[:-1], count, color='g')
            count, bins = np.histogram(cross_sp, bins=100)
            ax.plot(bins[:-1], count, color='r')
            ax.axvline(l0)
            ax.axvline(l1)
            ax.set_title(f'merge {equal} shift {shift}')

            
            fig, ax = plt.subplots()
            ax.plot(centroid0.T.flatten())
            ax.plot(centroid1.T.flatten())
            ax.set_title(f'shift {shift}')
            
            plt.show()
            
        
        if equal:
            break
    
    return equal



#~ def nearest_neighbor_similarity(cc):
    
    #~ labels = cc.positive_cluster_labels
    #~ n = len(labels)
    
    #~ snn = np.zeros((n,n), dtype='float32')
    
    #~ for i, label0 in enumerate(labels):
        #~ for j, label1 in enumerate(labels):
            #~ if j<i:
                #~ continue
            
            #~ waveforms0 = cc.get_cached_waveforms(label0)
            #~ waveforms1 = cc.get_cached_waveforms(label1)
            #~ snn[i, j] = one_nearest_neighbor_similarity(waveforms0, waveforms1, n_neighbors=10)
            
            #~ print('i', i, 'j', j, 'snn',  snn[i, j])
            
    
    #~ return snn


#~ def one_nearest_neighbor_similarity(waveforms0, waveforms1, n_neighbors=10):
        
    #~ from sklearn.neighbors import NearestNeighbors
    
    #~ n0 = waveforms0.shape[0]
    
    #~ waveforms = np.concatenate((waveforms0, waveforms1), axis=0)
    #~ n = waveforms.shape[0]
    #~ wf_flat = waveforms.reshape(n,-1)
    
    #~ nn = NearestNeighbors(n_neighbors=n_neighbors)
    #~ nn.fit(wf_flat)
    #~ distances, indices = nn.kneighbors(wf_flat)
    #~ nn0 = np.sum((indices[:n0] < n0).astype(int), axis=1)
    #~ nn1 = np.sum((indices[n0:] >= n0).astype(int), axis=1)
    #~ one_snn = np.mean(n_neighbors - np.concatenate((nn0, nn1))) / n_neighbors
    
    #~ return one_snn
    



#~ def get_normalized_centroids(centroids, peak_sign, sparse_thresh=None):
    #~ nb_clus, width, nb_chan = centroids.shape
    
    #~ normalized_centroids = np.zeros_like(centroids)
    
    #~ for cluster_idx in range(nb_clus):
        #~ centroid = centroids[cluster_idx, :, :]
        #~ centroid_normalized = centroid.copy()
        
        #~ if sparse_thresh is not None:
            #~ mask = np.any(np.abs(centroid) > sparse_thresh, axis=0)
            #~ centroid_normalized[:, ~mask] = 0
            
        #~ if peak_sign == '-':
            #~ centroid_normalized /= np.abs(np.min(centroid_normalized))
        #~ elif peak_sign == '+':
            #~ centroid_normalized /= np.abs(np.max(centroid_normalized))
        #~ centroid_normalized /= np.sum(centroid_normalized**2)
        #~ normalized_centroids[cluster_idx, :] = centroid_normalized

    #~ return normalized_centroids


def compute_sparse_mask(centroids, mode, method='thresh', thresh=None, nbest=None):
    nb_clus, width, nb_chan = centroids.shape
    
    sparse_mask_level = np.ones((nb_clus, nb_chan), dtype='bool')
    
    if mode == 'sparse':
        if method == 'thresh':
            for cluster_idx in range(nb_clus):
                centroid = centroids[cluster_idx, :, :]
                mask = np.any(np.abs(centroid) > thresh, axis=0)
                sparse_mask_level[cluster_idx, :] = mask
        elif method == 'nbest':
            for cluster_idx in range(nb_clus):
                centroid = centroids[cluster_idx, :, :]
                amplitudes = np.max(np.abs(centroid), axis=0)
                order = np.argsort(amplitudes)[::-1]
                best_channels = order[:nbest]
                sparse_mask_level[cluster_idx, :] = False
                sparse_mask_level[cluster_idx, :][best_channels] = True
                
    
    return sparse_mask_level

def compute_shared_channel_mask(centroids, mode, sparse_thresh):
    n = centroids.shape[0]
    
    share_channel_mask =np.ones((n, n), dtype='bool')
    
    if mode == 'dense':
        return share_channel_mask

    sparse_mask = compute_sparse_mask(centroids, mode, method='thresh', thresh=sparse_thresh)
    
    for cluster_idx0 in range(n):
        
        #~ print('cluster_idx0', cluster_idx0)
        
        centroid0 = centroids[cluster_idx0, :, :]
        
        mask0 = sparse_mask[cluster_idx0]
        n0 = np.sum(mask0)
        
        for cluster_idx1 in range(n):
            mask1 = sparse_mask[cluster_idx1]
            nshared = np.sum(mask0 & mask1)
            
            n1 = np.sum(mask1)
            
            m = min(n0, n1)
            if m <= 4:
                if nshared == 0:
                    channel_shared = False
                else:
                    channel_shared = True
            else:
                if nshared > m * 0.5:
                    channel_shared = True
                else:
                    channel_shared = False
            
            #~ print('channel_shared', channel_shared)
            #~ fig, ax = plt.subplots()
            #~ ax.plot(centroid0.T.flatten())
            #~ centroid1 = centroids[cluster_idx1, :, :]
            #~ ax.plot(centroid1.T.flatten())
            #~ ax.set_title(f'{cluster_idx0} {cluster_idx1} channel_shared{channel_shared}')
            #~ plt.show()
            
            
            share_channel_mask[cluster_idx0, cluster_idx1] = channel_shared

    return share_channel_mask




def compute_projection(centroids, sparse_mask_level1):

    n = centroids.shape[0]
    #~ print('n', n)
    
    #~ flat_shape = n, centroids.shape[1] * centroids.shape[2]
    projections = np.zeros(centroids.shape, dtype='float32')
    
    neighbors = {}
    
    for cluster_idx0 in range(n):
    
        # only centroids / channel that are on sparse mask
        chan_mask = sparse_mask_level1[cluster_idx0, :]
        
        # case1 sharing chan_mask
        #~ nover = np.sum(sparse_mask_level1[:, chan_mask], axis=1)
        #~ clus_mask = (nover > 0)
        
        # case2 5 nearest template
        #~ distances = np.sum((centroids - centroids[[cluster_idx0], :, :])**2, axis=(1,2))
        #~ order = np.argsort(distances)
        #~ nearest = order[:8]
        #~ clus_mask = np.zeros(n, dtype='bool')
        #~ clus_mask[nearest] = True
        
        # case3 all cluster
        clus_mask = np.ones(n, dtype='bool')
        
        # case 4 : sharing chan_mask + N nearest
        #~ nover = np.sum(sparse_mask_level1[:, chan_mask], axis=1)
        #~ sharing_mask = (nover > 0)
        #~ distances = np.sum((centroids - centroids[[cluster_idx0], :, :])**2, axis=(1,2))
        #~ distances[~sharing_mask] = np.inf
        #~ order = np.argsort(distances)
        #~ nearest = order[:8]
        #~ clus_mask = np.zeros(n, dtype='bool')
        #~ clus_mask[nearest] = True
        
        local_indexes0, = np.nonzero(clus_mask)
        local_idx0 = np.nonzero(local_indexes0 == cluster_idx0)[0][0]
            
        local_nclus = np.sum(clus_mask)
        local_chan = np.sum(chan_mask)
        flat_centroids = centroids[clus_mask, :, :][:, :, chan_mask].reshape(local_nclus, -1).T.copy()
        #~ print('local_nclus', local_nclus, 'local_chan', local_chan)
        
        
        flat_centroid0 = flat_centroids[:, local_idx0]
        #~ flat_centroid0 = centroids[cluster_idx0, :, :][:, chan_mask].flatten()
        
        other_mask = np.ones(local_nclus, dtype='bool')
        other_mask[local_idx0] = False
        
        if np.sum(other_mask) > 0:
            other_centroids = flat_centroids[:, other_mask]

            ind_min = np.argmin(np.sum((other_centroids - flat_centroid0[:, np.newaxis])**2, axis=0))
            other_select = [ind_min]
            
            
            while True:
                if len(other_select) == 1:
                    centroid0_proj = other_centroids[:, other_select[0]]
                else:
                    # This find point in orthogonal hyperplan to the centroid
                    centerred_other_centroids = other_centroids[:, other_select].copy()
                    shift = -centerred_other_centroids[:, 0]
                    centerred_other_centroids += shift[:, np.newaxis]
                    centerred_other_centroids = centerred_other_centroids[:, 1:]
                    u,s,vh = scipy.linalg.svd(centerred_other_centroids, full_matrices=False)
                    centroid0_proj = u @ u.T @ (flat_centroid0 + shift) - shift
                
                local_projector = centroid0_proj - flat_centroid0
                local_projector /= np.sum(local_projector**2)                

                # test if noise (0,0,0, ...) must be in the hyperplane
                noise_feat = (- flat_centroid0).T @ local_projector
                if np.abs(noise_feat) < 1.:
                    centerred_other_centroids = other_centroids[:, other_select].copy()
                    u,s,vh = scipy.linalg.svd(centerred_other_centroids, full_matrices=False)
                    centroid0_proj = u @ u.T @ (flat_centroid0)
                    local_projector = centroid0_proj - flat_centroid0
                    local_projector /= np.sum(local_projector**2)
                    #~ print('cluster_idx0', cluster_idx0, 'WITH noise')
                #~ else:
                    #~ print('cluster_idx0', cluster_idx0, 'WITHOUT noise')
                    
                other_feat = (other_centroids - flat_centroid0[:, np.newaxis]).T @ local_projector
                
                ind,  = np.nonzero(np.abs(other_feat) < 1.)
                
                if ind.size == 0:
                    break
                else:
                    # TODO : try to add them one by one and not sevral at each loop
                    not_in = np.ones(other_feat.size, dtype='bool')
                    not_in[other_select] = False
                    too_small = (np.abs(other_feat) < 1.)
                    others_candidate, = np.nonzero(not_in & too_small)
                    if others_candidate.size ==0:
                        break
                    smallest_ind = np.argmin(np.abs(other_feat[others_candidate]))
                    other_select.append(others_candidate[smallest_ind])
                    #~ print(cluster_idx0, 'other_select', other_select)
                
            neighbors[cluster_idx0] = np.nonzero(other_mask)[0][other_select]
            ortho_complement = local_projector
        else:
            # alone one theses channels = make projection with noise
            ortho_complement = 0 - flat_centroid0
            ortho_complement /= np.sum(ortho_complement**2)                
            
            neighbors[cluster_idx0] = {}
        
        #~ print('neighbors[cluster_idx0]', neighbors[cluster_idx0])

        
        
        projections[cluster_idx0, :, :][:, chan_mask] = ortho_complement.reshape(centroids.shape[1], local_chan)

    
    return projections, neighbors




def compute_boundaries(cc, centroids, sparse_mask_level1, projections, neighbors, plot_debug=False):
    #~ n = len(cluster_labels)
    n = centroids.shape[0]

    keep = cc.cluster_labels>=0
    cluster_labels = cc.clusters[keep]['cluster_label'].copy()

    
    scalar_products = np.zeros((n+1, n+1), dtype=object)
    boundaries = np.zeros((n, 4), dtype='float32')
    
    # compute all scalar product with projection
    for cluster_idx0, label0 in enumerate(cluster_labels):
        chan_mask = sparse_mask_level1[cluster_idx0, :]
        projector =projections[cluster_idx0, :][:, chan_mask].flatten()
        
        wf0 = cc.get_cached_waveforms(label0)
        flat_centroid0 = centroids[cluster_idx0, :, :][:, chan_mask].flatten()
        #~ print('cluster_idx0',cluster_idx0, wf0.shape)
        
        wf0 = wf0[:, :, chan_mask].copy()
        flat_wf0 = wf0.reshape(wf0.shape[0], -1)
        feat_wf0 = (flat_wf0 - flat_centroid0) @ projector
        feat_centroid0 = (flat_centroid0 - flat_centroid0) @ projector
        
        scalar_products[cluster_idx0, cluster_idx0] = feat_wf0
        
        for cluster_idx1, label1 in enumerate(cluster_labels):
            if cluster_idx0 == cluster_idx1:
                continue
            
            centroid1 = centroids[cluster_idx1, :, :][:, chan_mask]
            wf1 = cc.get_cached_waveforms(label1)
            wf1 = wf1[:, :, chan_mask].copy()
            flat_wf1 = wf1.reshape(wf1.shape[0], -1) 
            feat_centroid1 = (centroid1.flatten() - flat_centroid0) @ projector
            feat_wf1 = (flat_wf1- flat_centroid0) @ projector
            
            scalar_products[cluster_idx0, cluster_idx1] = feat_wf1
        
        noise = cc.some_noise_snippet
        noise = noise[:, :, chan_mask].copy()
        flat_noise = noise.reshape(noise.shape[0], -1)
        feat_noise = (flat_noise - flat_centroid0) @ projector
        scalar_products[cluster_idx0, -1] = feat_noise
    
    # find boudaries
    for cluster_idx0, label0 in enumerate(cluster_labels):
        inner_sp = scalar_products[cluster_idx0, cluster_idx0]
        med, mad = median_mad(inner_sp)
        
        mad_factor = 6
        low = med - mad_factor * mad
        initial_low = low
        high = med + mad_factor * mad
        initial_high = high


        # optimze boudaries with accuracy
        high_clust = []
        high_lim = []
        low_clust = []
        low_lim = []
        for cluster_idx1, label1 in enumerate(cluster_labels):
            # select dangerous cluster
            if cluster_idx1 == cluster_idx0:
                continue
            cross_sp = scalar_products[cluster_idx0, cluster_idx1]
            med, mad = median_mad(cross_sp)
            if med > high and (med - mad_factor * mad) < high :
                high_clust.append(cluster_idx1)
                high_lim.append(med - mad_factor * mad)
            if med < low and (med + mad_factor * mad) > low:
                low_clust.append(cluster_idx1)
                low_lim.append(med + mad_factor * mad)

        noise_sp = scalar_products[cluster_idx0, -1]
        med, mad = median_mad(noise_sp)
        if med > high and (med - mad_factor * mad) < high :
            high_clust.append(-1)
            high_lim.append(med - mad_factor * mad)
        # TODO if noise in low limits
        #~ print()
        #~ print('initial_low', initial_low)
        #~ print('initial_high', initial_high)
        #~ print('high_clust', high_clust, 'high_lim', high_lim)
        #~ print('low_clust', low_clust, 'low_lim', low_lim)

        if len(high_clust) > 0:
            l0 = min(high_lim)
            l1 = initial_high
            step = (l1-l0)/20.
            
            all_sp = np.concatenate([scalar_products[cluster_idx0, idx1] for idx1 in high_clust])
            limits = np.arange(l0, l1, step) 
            accuracies = []
            for l in limits:
                tp = scalar_products[cluster_idx0, cluster_idx0].size
                fn = np.sum(scalar_products[cluster_idx0, cluster_idx0] > l)
                fp = np.sum(all_sp < l)
                accuracy = tp / (tp + fn + fp)
                accuracies.append(accuracy)
            best_lim = limits[np.argmax(accuracies)]
            boundaries[cluster_idx0, 1] = min(best_lim, 0.5)
            
            #~ fig, ax = plt.subplots()
            #~ ax.plot(limits, accuracies)
            #~ ax.axvline(best_lim)
            #~ ax.set_title(f'high {cluster_idx0}')
            #~ plt.show()
            
        else:
            boundaries[cluster_idx0, 1] = min(initial_high, 0.5)
        
        if len(low_clust) > 0:
            l1 = max(low_lim)
            l0 = initial_low
            step = (l1-l0)/20.
            
            all_sp = np.concatenate([scalar_products[cluster_idx0, idx1] for idx1 in low_clust])
            limits = np.arange(l0, l1, step) 
            accuracies = []
            for l in limits:
                tp = scalar_products[cluster_idx0, cluster_idx0].size
                fn = np.sum(scalar_products[cluster_idx0, cluster_idx0] <l)
                fp = np.sum(all_sp > l)
                accuracy = tp / (tp + fn + fp)
                accuracies.append(accuracy)
            best_lim = limits[np.argmax(accuracies)]
            boundaries[cluster_idx0, 0] = max(best_lim, -0.5)
        
        else:
            boundaries[cluster_idx0, 0] = max(initial_low, -0.5)

        
        boundaries[cluster_idx0, 2] = max(initial_low, -0.5)
        boundaries[cluster_idx0, 3] = min(initial_high, 0.5)                    
    
    #~ if plot_debug:
        #~ import matplotlib.pyplot as plt
        
        
    
    
    return boundaries, scalar_products






import numpy as np
import time
import os
import shutil
from pprint import pprint

from tridesclous import download_dataset
from tridesclous.dataio import DataIO
from tridesclous.catalogueconstructor import CatalogueConstructor
from tridesclous.autoparams import get_auto_params_for_catalogue
from tridesclous.cataloguetools import  apply_all_catalogue_steps

from matplotlib import pyplot as plt

    
    
    
def test_apply_all_catalogue_steps():
    if os.path.exists('test_cataloguetools'):
        shutil.rmtree('test_cataloguetools')
        
    dataio = DataIO(dirname='test_cataloguetools')
    #~ localdir, filenames, params = download_dataset(name='olfactory_bulb')
    localdir, filenames, params = download_dataset(name='locust')
    dataio.set_data_source(type='RawData', filenames=filenames, **params)
    
    params = get_auto_params_for_catalogue(dataio)
    
    cc = CatalogueConstructor(dataio, chan_grp=0)
    apply_all_catalogue_steps(cc, params, verbose=True)
    
    
    
    
    
if __name__ == '__main__':
    test_apply_all_catalogue_steps()
    
    


import pandas as pd
import numpy as np
from tridesclous.tools import *
from tridesclous.probe_list import probe_list

from urllib.request import urlretrieve
import time

def test_get_median_mad():
    pass
    

def test_FifoBuffer():
    n = 5
    fifo = FifoBuffer((1024+64, n), dtype='int16')
    for i in range(4):
        data = np.tile(np.arange(1024)[:, None], (1, n))+i*1024
        fifo.new_chunk(data, 1024*(i+1))
    data2 = fifo.get_data(3008, 4096)
    assert np.all(data2[0,:]==3008)
    assert np.all(data2[-1,:]==4095)
    #~ print(data2)

def test_get_neighborhood():
    geometry = [[0,0], [0, 100], [100, 100]]
    radius_um = 120
    n = get_neighborhood(geometry, radius_um)
    print(n)


def test_fix_prb_file_py2():
    distantfile = 'https://raw.githubusercontent.com/spyking-circus/spyking-circus/master/probes/kampff_128.prb'
    prb_filename = 'kampff_128.prb'
    urlretrieve(distantfile, prb_filename)

    fix_prb_file_py2(prb_filename)


def re_construct_probe_list():
    construct_probe_list()

def show_construct_probe_list():
    for k, v in probe_list.items():
        print('*'*20)
        print('**', k)
        print('*'*20)
        for e in v:
            print(e)


def test_compute_cross_correlograms():

    n = 1000000

    sample_rate=10000.
    bin_size=0.001
    window_size=0.2
    #~ symmetrize=True
    
    spike_times = np.random.rand(n)*10000
    spike_times.sort()
    spike_indexes = (spike_times*sample_rate).astype('int64')
    
    spike_labels = np.random.randint(0,6, size=n)
    
    spike_segments = np.zeros(n, dtype='int64')
    spike_segments[n//2] = 1
    
    cluster_labels = np.unique(spike_labels)
    

    t0 = time.perf_counter()
    cgc, bins = compute_cross_correlograms(spike_indexes, spike_labels, spike_segments,
                cluster_labels, sample_rate, window_size, bin_size)
    t1 = time.perf_counter()
    
    print(cgc.shape)
    print('compute_cross_correlograms', t1-t0)


def test_int32_to_rgba():
    r, g, b, a = int32_to_rgba(2654789321)
    print(r, g, b, a)
    
def test_rgba_to_int32():
    v = rgba_to_int32(158,60,222, 201)
    print(v)
    

if __name__ == '__main__':
    #~ test_get_median_mad()
    #~ test_FifoBuffer()
    #~ test_get_neighborhood()
    #~ test_fix_prb_file_py2()
    re_construct_probe_list()
    #~ show_construct_probe_list()
    #~ test_compute_cross_correlograms()
    #~ test_int32_to_rgba()
    #~ test_rgba_to_int32()
    




LABEL_TRASH = -1
LABEL_NOISE = -2
LABEL_ALIEN = -9
LABEL_UNCLASSIFIED = -10
LABEL_NO_WAVEFORM = -11




to_name = { -1: 'Trash', 
    -2 : 'Noise',
    -9: 'Alien',
    -10: 'Unclassified',
    -11: 'No waveforms',
}





from collections import OrderedDict

try:
    import pyopencl
    mf = pyopencl.mem_flags
    HAVE_PYOPENCL = True
except ImportError:
    HAVE_PYOPENCL = False
    

default_platform_index = None
default_device_index = None



class OpenCL_Helper:
    """
    Used in Peeler and SignalPreprocessor_OpenCL.
    
    Make possible to select device/platform with 3 flavor:
       * default at tridesclous level when default_platform_index/default_device_index is not None
       * at each Peeler.set_params(cl_platform_index=XXX, cl_device_index=YYY) to change this for each instance (make compraison)
       * default at pyopencl level when default_platform_index is None
    
    """

    def initialize_opencl(self, cl_platform_index=None, cl_device_index=None, ctx=None, queue=None):
        assert HAVE_PYOPENCL, 'PyOpenCL is not installed'

        global default_platform_index
        global default_device_index
        
        if ctx is None and queue is None:
            if cl_platform_index is None:
                if default_platform_index is not None and default_device_index is not None:
                    self.cl_platform_index = default_platform_index
                    self.cl_device_index = default_device_index
                    self.devices = [pyopencl.get_platforms()[self.cl_platform_index].get_devices()[self.cl_device_index]]
                    self.ctx = pyopencl.Context(self.devices)
                else:
                    self.ctx = pyopencl.create_some_context(interactive=False)
            else:
                self.cl_platform_index = cl_platform_index
                self.cl_device_index = cl_device_index
                self.devices = [pyopencl.get_platforms()[self.cl_platform_index].get_devices()[self.cl_device_index]]
                self.ctx = pyopencl.Context(self.devices)
            self.queue = pyopencl.CommandQueue(self.ctx)
        else:
            assert cl_platform_index is None and cl_device_index is None
            self.ctx = ctx
            self.queue = queue
        
        self.max_wg_size = self.ctx.devices[0].get_info(pyopencl.device_info.MAX_WORK_GROUP_SIZE)
        


def get_cl_device_list():
    
    device_indexes = []
    for platform_index, platform in enumerate(pyopencl.get_platforms()):
        for device_index,device in enumerate(platform.get_devices()):
            device_indexes.append((device.name, platform_index, device_index))
    
    return device_indexes


def set_default_cl_device(platform_index=None, device_index=None):
    global default_platform_index
    global default_device_index
    
    default_platform_index = platform_index
    default_device_index = device_index





import pytest
import os, tempfile, shutil
import numpy as np

from urllib.request import urlopen

from tridesclous import download_dataset
#~ from tridesclous import DataIO
from tridesclous.datasource import InMemoryDataSource, data_source_classes




def test_InMemoryDataSource():
    
    nparrays = [np.random.randn(10000, 2) for i in range(5) ]
    datasource = InMemoryDataSource(nparrays=nparrays, sample_rate=5000.)
    assert datasource.total_channel == 2
    assert datasource.sample_rate == 5000.
    assert datasource.nb_segment==5
    assert datasource.get_segment_shape(0) == (10000, 2)
    data = datasource.get_signals_chunk(seg_num=0)
    assert data.shape==datasource.get_segment_shape(0)
    

def test_RawDataSource():
    from tridesclous.datasource import RawDataSource
    
    localdir, filenames, params = download_dataset(name='olfactory_bulb')
    datasource = RawDataSource(filenames=filenames, **params)
    
    assert datasource.total_channel == 16
    assert datasource.sample_rate == 10000.
    assert datasource.nb_segment==3
    assert datasource.get_segment_shape(0) == (150000, 16)
    data = datasource.get_signals_chunk(seg_num=0)
    assert data.shape==datasource.get_segment_shape(0)


def test_NeoRawIOAggregator():
    
    url_for_tests = "https://web.gin.g-node.org/NeuralEnsemble/ephy_testing_data/raw/master/"
    distantfile = url_for_tests+'blackrock/FileSpec2.3001.ns5'
    
    localfile = 'FileSpec2.3001.ns5'
    
    if not os.path.exists(localfile):
        dist = urlopen(distantfile)
        with open(localfile, 'wb') as f:
            f.write(dist.read())
    
    
    from tridesclous import BlackrockDataSource
    
    datasource = BlackrockDataSource(filenames=[localfile, localfile])
    print(datasource.total_channel)
    print(datasource.nb_segment)
    print(datasource.get_segment_shape(0), datasource.get_segment_shape(1))
    assert datasource.get_segment_shape(0)==datasource.get_segment_shape(1)
    print(datasource.get_channel_names())
    data = datasource.get_signals_chunk(seg_num=0)
    assert data.shape==datasource.get_segment_shape(0)

    



if __name__=='__main__':
    test_InMemoryDataSource()
    test_RawDataSource()
    test_NeoRawIOAggregator()
    


"""
Some helping function that apply on catalogueconstructor (apply steps, sumamry, ...)

"""
import time
import copy
from pprint import pprint

from .cltools import HAVE_PYOPENCL
from .cluster import HAVE_ISOSPLIT5

try:
    import numba
    HAVE_NUMBA = True
except ImportError:
    HAVE_NUMBA = False
    

#TODO debug this when no peak at all
def apply_all_catalogue_steps(catalogueconstructor, params, verbose=True):
    """
    Helper function to call seuquentialy all catalogue steps with dict of params as input.
    Used by offline mainwinwod and OnlineWindow.
    Also used by spikeinterface.sorters.
    
    
    Usage:
      
      cc = CatalogueConstructor(dataio, chan_grp=0)
      params = get_auto_params_for_catalogue(dataio, chan_grp=0)
      apply_all_catalogue_steps(cc, params)
    """
    #~ if verbose:
        #~ print('apply all catalogue steps')
        #~ pprint(params)
    
    #~ pprint(params)
    
    cc = catalogueconstructor
    
    # global params
    d = {k:params[k] for k in ('chunksize', 'sparse_threshold', 'mode', 'memory_mode', 'n_jobs', 'n_spike_for_centroid', ) if k in params}
    
    cc.set_global_params(**d)
    
    # params preprocessor
    d = dict(params['preprocessor'])
    cc.set_preprocessor_params(**d)
    
    # params peak detector
    d = dict(params['peak_detector'])
    cc.set_peak_detector_params(**d)
    
    dataio = cc.dataio
    
    #TODO offer noise esatimation duration somewhere
    noise_duration = min(10., params['duration'], dataio.get_segment_length(seg_num=0)/dataio.sample_rate*.99)
    #~ print('noise_duration', noise_duration)
    t1 = time.perf_counter()
    cc.estimate_signals_noise(seg_num=0, duration=noise_duration)
    t2 = time.perf_counter()
    if verbose:
        print('estimate_signals_noise', t2-t1)
    
    t1 = time.perf_counter()
    cc.run_signalprocessor(duration=params['duration'])
    t2 = time.perf_counter()
    if verbose:
        print('run_signalprocessor', t2-t1)
    
    #~ t1 = time.perf_counter()
    #~ cc.extract_some_waveforms(**params['extract_waveforms'], recompute_all_centroid=False)
    #~ t2 = time.perf_counter()
    #~ if verbose:
        #~ print('extract_some_waveforms', t2-t1)
    
    #~ t1 = time.perf_counter()
    #~ cc.clean_waveforms(**params['clean_waveforms'], recompute_all_centroid=False)
    #~ t2 = time.perf_counter()
    #~ if verbose:
        #~ print('clean_waveforms', t2-t1)
    
    #~ t1 = time.perf_counter()
    #~ n_left, n_right = cc.find_good_limits(mad_threshold = 1.1,)
    #~ t2 = time.perf_counter()
    #~ print('find_good_limits', t2-t1)

    cc.set_waveform_extractor_params(**params['extract_waveforms'])
    
    t1 = time.perf_counter()
    cc.clean_peaks(**params['clean_peaks'])
    t2 = time.perf_counter()
    if verbose:
        print('clean_peaks', t2-t1)
    
    t1 = time.perf_counter()
    cc.sample_some_peaks(**params['peak_sampler'])
    t2 = time.perf_counter()
    if verbose:
        print('sample_some_peaks', t2-t1)
    
    t1 = time.perf_counter()
    cc.extract_some_noise(**params['noise_snippet'])
    t2 = time.perf_counter()
    if verbose:
        print('extract_some_noise', t2-t1)
    
    #~ print(cc)
    
    t1 = time.perf_counter()
    cc.extract_some_features(method=params['feature_method'], **params['feature_kargs'])
    t2 = time.perf_counter()
    if verbose:
        print('extract_some_features', t2-t1)
    
    #~ print(cc)
    
    t1 = time.perf_counter()
    cc.find_clusters(method=params['cluster_method'], recompute_centroid=False, order=False, **params['cluster_kargs'])
    t2 = time.perf_counter()
    if verbose:
        print('find_clusters', t2-t1)
    
    t1 = time.perf_counter()
    cc.cache_some_waveforms()
    t2 = time.perf_counter()
    if verbose:
        print('cache_some_waveforms', t2-t1)
    
    t1 = time.perf_counter()
    cc.compute_all_centroid()
    t2 = time.perf_counter()
    if verbose:
        print('compute_all_centroid', t2-t1)

    #~ cc.order_clusters(by='waveforms_rms')
    #~ t2 = time.perf_counter()
    #~ if verbose:
        #~ print('order_clusters', t2-t1)

    # clean cluster steps
    pclean = params['clean_cluster']
    
    if pclean['apply_auto_split']:
        t1 = time.perf_counter()
        cc.auto_split_cluster()
        if verbose:
            t2 = time.perf_counter()
            print('auto_split_cluster', t2-t1)
    
    if pclean['apply_trash_not_aligned']:
        t1 = time.perf_counter()
        cc.trash_not_aligned()
        if verbose:
            t2 = time.perf_counter()
            print('trash_not_aligned', t2-t1)
    
    if pclean['apply_auto_merge_cluster']:
        t1 = time.perf_counter()
        cc.auto_merge_cluster()
        if verbose:
            t2 = time.perf_counter()
            print('auto_merge_cluster', t2-t1)
    
    if pclean['apply_trash_low_extremum']:
        t1 = time.perf_counter()
        cc.trash_low_extremum()
        if verbose:
            t2 = time.perf_counter()
            print('trash_low_extremum', t2-t1)

    if pclean['apply_trash_small_cluster']:
        t1 = time.perf_counter()
        cc.trash_small_cluster()
        if verbose:
            t2 = time.perf_counter()
            print('trash_small_cluster', t2-t1)
    
    cc.order_clusters(by='waveforms_rms')
    
    t1 = time.perf_counter()
    cc.make_catalogue_for_peeler(**params['make_catalogue'])
    t2 = time.perf_counter()
    if verbose:
        print('make_catalogue_for_peeler', t2-t1)
    




try:
    import numba
    from numba import jit, prange
    HAVE_NUMBA = True
except ImportError:
    HAVE_NUMBA = False

import numpy as np

# OLD
@jit(parallel=True)
def numba_loop_sparse_dist(waveform, centers,  mask):
    nb_clus, width, nb_chan = centers.shape
    
    rms_waveform_channel = np.sum(waveform**2, axis=0)#.astype('float32')
    waveform_distance = np.zeros((nb_clus,), dtype=np.float32)
    
    for clus in range(nb_clus):
        sum = 0
        for c in range(nb_chan):
            if mask[clus, c]:
                for s in range(width):
                    d = waveform[s, c] - centers[clus, s, c]
                    sum += d*d
            else:
                sum +=rms_waveform_channel[c]
        waveform_distance[clus] = sum
    
    return waveform_distance



@jit(parallel=True)
def numba_sparse_scalar_product(fifo_residuals, left_ind, centers, projector, peak_chan_ind,
                        sparse_mask_level1, ):
    nb_clus, width, nb_chan = centers.shape
    
    scalar_product = np.zeros((nb_clus,), dtype=np.float32)
    
    for clus_idx in prange(nb_clus):
        if not sparse_mask_level1[clus_idx, peak_chan_ind]:
            scalar_product[clus_idx] = 10. # equivalent to np.inf
        else:
            sum_sp = 0.
            for chan in range(nb_chan):
                if sparse_mask_level1[clus_idx, chan]:
                    for s in range(width):
                        v = fifo_residuals[left_ind+s, chan]
                        ct = centers[clus_idx, s, chan]
                        w = projector[clus_idx, s, chan]
                        
                        sum_sp += (v - ct) * w
                    
            scalar_product[clus_idx] = sum_sp
    
    return scalar_product
    





@jit(parallel=True)
def numba_explore_best_shift(fifo_residuals, left_ind, centers, projector, candidates_idx,  maximum_jitter_shift, common_mask, sparse_mask_level1):

    nb_clus, width, nb_chan = centers.shape
    
    n_shift = maximum_jitter_shift*2 +1
    n_clus = len(candidates_idx)
    
    shift_scalar_product = np.zeros((n_clus, n_shift), dtype=np.float32)
    shift_distance = np.zeros((n_clus, n_shift), dtype=np.float32)
    
    for shi in prange(n_shift):
        shift =  shi - maximum_jitter_shift
        
        for clus in prange(len(candidates_idx)):
            clus_idx = candidates_idx[clus]
            
            sum_sp = 0.
            sum_d = 0.
            for chan in range(nb_chan):
                if common_mask[chan] or sparse_mask_level1[clus_idx, chan]:
                    for s in range(width):
                        v = fifo_residuals[left_ind+s+shift, chan]
                        ct = centers[clus_idx, s, chan]
                        
                        if sparse_mask_level1[clus_idx, chan]:
                            sum_sp += (v - ct) * projector[clus_idx, s, chan]
                        
                        if common_mask[chan]:
                            sum_d += (v - ct) * (v - ct)
                    
            shift_scalar_product[clus, shi] = sum_sp
            shift_distance[clus, shi] = sum_d
    
    return shift_scalar_product, shift_distance





@jit(parallel=True)
def peak_loop_plus(sigs, sig_center, mask_peaks, n_span, thresh, peak_sign, neighbours):
    for chan in prange(sig_center.shape[1]):
        for s in range(mask_peaks.shape[0]):
            if not mask_peaks[s, chan]:
                continue
            for neighbour in neighbours[chan, :]:
                if neighbour<0:
                    continue
                for i in range(n_span):
                    if chan != neighbour:
                        mask_peaks[s, chan] &= sig_center[s, chan] >= sig_center[s, neighbour]
                    mask_peaks[s, chan] &= sig_center[s, chan] > sigs[s+i, neighbour]
                    mask_peaks[s, chan] &= sig_center[s, chan]>=sigs[n_span+s+i+1, neighbour]
                    if not mask_peaks[s, chan]:
                        break
                if not mask_peaks[s, chan]:
                    break
    return mask_peaks


@jit(parallel=True)
def peak_loop_minus(sigs, sig_center, mask_peaks, n_span, thresh, peak_sign, neighbours):
    for chan in prange(sig_center.shape[1]):
        for s in range(mask_peaks.shape[0]):
            if not mask_peaks[s, chan]:
                continue
            for neighbour in neighbours[chan, :]:
                if neighbour<0:
                    continue
                for i in range(n_span):
                    if chan != neighbour:
                        mask_peaks[s, chan] &= sig_center[s, chan] <= sig_center[s, neighbour]
                    mask_peaks[s, chan] &= sig_center[s, chan] < sigs[s+i, neighbour]
                    mask_peaks[s, chan] &= sig_center[s, chan]<=sigs[n_span+s+i+1, neighbour]                        
                    if not mask_peaks[s, chan]:
                        break
                if not mask_peaks[s, chan]:
                    break
    return mask_peaks


def numba_get_mask_spatiotemporal_peaks(sigs, n_span, thresh, peak_sign, neighbours):
    sig_center = sigs[n_span:-n_span, :]

    if peak_sign == '+':
        mask_peaks = sig_center>thresh
        mask_peaks = peak_loop_plus(sigs, sig_center, mask_peaks, n_span, thresh, peak_sign, neighbours)

    elif peak_sign == '-':
        mask_peaks = sig_center<-thresh
        mask_peaks = peak_loop_minus(sigs, sig_center, mask_peaks, n_span, thresh, peak_sign, neighbours)

    return mask_peaks




